Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path.
Require Import pred prelude idynamic ordtype finmap pcm coding unionmap heap.
(* Require Import feap state zigzag world inject rely hide atomic autosplit. *)
(* Require Import privstate footprint model lemmas. *)
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(* Last_key and Fresh are constructs that work *)
(* for any union map with integer keys *)
(* Should be developed more generally for any union map class *)
(* with a proof that key U = nat, but I can't bother right now *)

Section FreshLastKey.
Variable V : Type.
Implicit Type f : union_map [ordType of nat] V. 

Definition last_key f := last 0 (keys_of f).

Lemma last_key0 : last_key Unit = 0.
Proof. by rewrite /last_key /Unit /= !umE. Qed.

Lemma last_key_dom f : valid f -> last_key f \notin dom f -> f = Unit. 
Proof.
rewrite /valid /= /last_key /Unit /= !umE /= -{4}[f]UMC.tfE.
case: (UMC.from f)=>//=; case=>s H /= H1 _ /seq_last_in. 
rewrite /UM.empty UMC.eqE UM.umapE /supp fmapE /= {H H1}.
by elim: s.
Qed.

Lemma last_key_max f x : x \in dom f -> x <= last_key f.
Proof.
rewrite /last_key /= !umE; case: (UMC.from f)=>//; case=>s H _ /=. 
rewrite /supp /ord /= (leq_eqVlt x) orbC. 
by apply: sorted_last_key_max (sorted_oleq H).
Qed.

Lemma max_key_last f x : 
        x \in dom f -> {in dom f, forall y, y <= x} -> last_key f = x.
Proof.
rewrite /last_key !umE; case: (UMC.from f)=>//; case=>s H _ /=.
move=>H1 /= H2; apply: sorted_max_key_last (sorted_oleq H) H1 _. 
by move=>z /(H2 z); rewrite leq_eqVlt orbC. 
Qed.

Lemma last_keyPt (x : nat) v : last_key (x \\-> v) = x.
Proof. by rewrite /last_key /um_pts /= !umE. Qed.

Lemma hist_path f : path oleq 0 (keys_of f).
Proof.
rewrite !umE; case: (UMC.from f)=>// {f} f /= _; case: f; case=>//= x s H.
rewrite {1}/oleq /ord /= orbC -leq_eqVlt /=.
by apply: sub_path H=>z y; rewrite /oleq=>->. 
Qed.

Lemma last_key_mono f1 f2 : 
        {subset dom f1 <= dom f2} -> last_key f1 <= last_key f2.
Proof.
rewrite leq_eqVlt orbC=>H; apply: (@keys_last_mono _ _ _ f1 f2);
try by apply: hist_path.
by move=>x /=; move: (H x); rewrite -!keys_dom. 
Qed.

(* freshness *)

Definition fresh f := (last_key f).+1. 

Lemma dom_ordfresh f x : x \in dom f -> x < fresh f.
Proof. by move/last_key_max. Qed.

Lemma dom_freshn f n : fresh f + n \notin dom f.
Proof. by apply: contra (@dom_ordfresh _ _) _; rewrite -leqNgt leq_addr. Qed.

Lemma dom_fresh f : fresh f \notin dom f.
Proof. by move: (dom_freshn f 0); rewrite addn0.  Qed.

Lemma valid_fresh f v : valid (f \+ fresh f \\-> v) = valid f.
Proof. by rewrite joinC um_validPtUn dom_fresh andbT. Qed.

Lemma last_fresh f v : valid f -> last_key (f \+ fresh f \\-> v) = fresh f.
Proof. 
move=>Vf; apply: max_key_last=>[|x] /=.
- by rewrite domUn inE valid_fresh Vf um_domPt inE eq_refl orbT.
rewrite domUn inE /= valid_fresh Vf /= um_domPt inE /= orbC eq_sym.
by rewrite leq_eqVlt; case: eqP=>//= _; apply: dom_ordfresh.
Qed.

End FreshLastKey.

(* all of the above to be moved to unionmap.v *)



(*****************)
(* Now histories *)
(*****************)

Section Histories.
Variable S : set. (* snapshot type *)
Definition hist := union_map [ordType of nat] (prodSet S S).
Implicit Type h : hist.

Definition well_formed h := 
  (* Starts from a stuttering element *)
  (exists s, find 0 h = Some (s, s)) /\
  (* Subsequent snapshots overlap *)
  forall k, k \in dom h -> k.+1 \in dom h -> 
    exists s1 s s2, find k h = Some (s1, s) /\ find k.+1 h = Some (s, s2).

Lemma wf_zero h : well_formed h -> 0 \in dom h.
Proof. by case; case=>s /find_some. Qed. 

Lemma wf_valid h : well_formed h -> valid h.
Proof. by move/wf_zero /dom_valid. Qed.

Definition complete h := 
  well_formed h /\ forall k, k <= last_key h -> k \in dom h.

Lemma complete_valid (h: hist) : complete h -> valid h. 
Proof. by case=>/wf_valid. Qed.

Notation "'[' s 'in' h 'at' t ']'" := (exists s', find t h = Some (s', s)).
Notation "'[' s 'in' h ']'" := (exists t, [s in h at t]).

Definition cont (P: S -> S -> Prop) h: Prop :=
  forall t s1 s2, 
    [s1 in h at t] -> [s2 in h at t.+1] -> P s1 s2. 

Lemma fresh_wf h s s2: 
        well_formed h -> [s in h at last_key h] ->
        well_formed (h \+ fresh h \\-> (s, s2)).
Proof.
move=>Hw; case: (Hw); case=>sx Hs /= He [s1] Hf; split=>[|k].
- by exists sx; rewrite findUnL ?valid_fresh // ?wf_valid ?wf_zero.
case E: (k == last_key h).
- exists s1, s, s2; rewrite !findUnL ?valid_fresh ?wf_valid // (eqP E).
  by rewrite (find_some Hf) (negbTE (dom_fresh h)) um_findPt.
rewrite !domUn !inE /= valid_fresh wf_valid //= !um_domPt !inE /= eq_sym.
rewrite eqSS -(eq_sym k) E orbF -{1}addn1 orbC. 
case: eqP (dom_freshn h 1)=>[-> /negbTE ->|_ _] //=; rewrite addn1=>H1 H2.
by rewrite !findUnL ?valid_fresh ?wf_valid // H1 H2; apply: He.
Qed.

Lemma fresh_complete h s s2: 
        complete h -> [s in h at last_key h] ->
        complete (h \+ fresh h \\-> (s, s2)).
Proof. 
case=>W H1 H2. 
split=>[|k H3]; first by apply:(@fresh_wf h s s2 W H2).
case X: (k == fresh h).
- rewrite joinC um_domPtUn inE; apply/andP; split; 
  [by apply:wf_valid; rewrite joinC; apply: fresh_wf| by rewrite (eqP X); apply/orP; left].
- rewrite joinC um_domPtUn inE; apply/andP; split; first by apply:wf_valid; rewrite joinC; apply: fresh_wf.
apply/orP;right; apply:(H1). move:(wf_valid W)=>Vh; rewrite last_fresh in H3=>//. 
move: H3; rewrite leq_eqVlt; case/orP; first by move/eqP: X =>Q1 /eqP. 
by rewrite /fresh /= ltnS.
Qed.

Lemma fresh_cont P h s s2: 
  valid h -> [s in h at last_key h] -> cont P h -> P s s2 ->
  cont P (h \+ (fresh h) \\-> (s, s2)).
Proof.
move=>Vh Hl Hc Hp. case:(Hl)=>s' /find_some=>D.
move=> k t1 t2; case E: (k == last_key h).
- rewrite !findUnL ?valid_fresh // (eqP E) D=> H1 [s''].
  case:(Hl)=>?; case:(H1)=>?->; case=>_ Z; subst t1.
  move:(dom_fresh h); rewrite /fresh =>H2; rewrite (negbTE H2) um_findPt. 
  by case=>Z1 Z2; subst t2 s''. 
case E1: (k == fresh h); case=>z1 H1; case=>z2 H2.
- move/find_some: H2. rewrite joinC um_domPtUn !inE joinC ?valid_fresh.
  case/andP=>_. move/eqP: E1=>->; case/orP.
+ move/eqP=>E3. move:(ltnSn (fresh h)); rewrite -{2}E3. 
  by rewrite (ltnn (fresh h)).
+ by move: (dom_freshn h 1); rewrite addn1=>H2; rewrite (negbTE H2).
move: (H1)(H2); rewrite !findUnL ?valid_fresh //.
case:ifP=>Z1 F1; case:ifP=>Z2 F2. 
have Y1: [t1 in h at k] by exists z1. 
have Y2: [t2 in h at k.+1] by exists z2. 
by move:(Hc k t1 t2 Y1 Y2)=>{Y1 Y2}.
- move/find_some: F2; rewrite um_domPt !inE /fresh /= eq_sym eqSS.
  by move/eqP; move/eqP: E. 
- move/find_some: F1; rewrite um_domPt !inE /= eq_sym.
  by move/eqP; move/eqP: E1. 
- move/find_some: F1; rewrite um_domPt !inE /= eq_sym;
  by move/eqP; move/eqP: E1. 
Qed.

End Histories.

Notation "'[' s 'in' h 'at' t ']'" := (exists s', find t h = Some (s', s)).
Notation "'[' s 'in' h ']'" := (exists t, [s in h at t]).

(* explicitly moving by existential abstraction *)
Lemma find_in S t (h : hist S) s1 s2 : find t h = Some (s1, s2) -> [s2 in h at t].
Proof. by exists s1. Qed.

Lemma at_in S t (h : hist S) s : [s in h at t] -> [s in h].
Proof. by exists t. Qed.

Lemma last_key_in S x t (h : hist S) : [x in h at t] -> t <= last_key h.
Proof. by case=>? /find_some; apply: last_key_max. Qed.

(*************************************************)
(* two different kinds of stability of histories *)
(*************************************************)

Definition local_stable S (h : hist S) (p : S -> bool) := 
  forall t s' s, find t h = Some (s, s') -> p s -> p s'.

Definition global_stable S (h : hist S) (p : S -> bool) := 
  forall t s t' s', t <= t' -> [s in h at t] -> [s' in h at t'] -> p s -> p s'.

(* in complete histories, local stable implies global stable *)

(* some helper lemmas for the proof that local stability implies local one *)
(* probably should be embedded into the proof of complete_stable *)
Lemma complete_pred S (h : hist S) s t :
        complete h -> [s in h at t.+1] -> exists s', [s' in h at t].
Proof. 
by case=>_ H /last_key_in /ltnW /H; case: dom_find=>// [[v1 v2]] ->; eauto. 
Qed.

Lemma wf_agree S (h : hist S) s s' t: 
       well_formed h -> [s in h at t] -> [s' in h at t.+1] -> 
       find t.+1 h = Some (s, s').
Proof.
case=>_ H [x] E1 [x'] E2.
case: (H _ (find_some E1) (find_some E2))=>s1 [s2][s3][].
by rewrite E1 E2; case=>_ <- [<-].
Qed.

Lemma complete_stable S (h : hist S) (p : S -> bool) : 
        complete h -> local_stable h p -> global_stable h p. 
Proof.
move=>C LS t s t' s' /subnKC <- H. 
elim: {t'} (t' - t) s'=>[|m IH] s'.
- by rewrite addn0; case: H=> x' -> [?][_ <-]. 
rewrite addnS=>H' P; case: (complete_pred C H')=>s'' H''.
by apply: LS (wf_agree (proj1 C) H'' H') (IH _ H'' P).
Qed.

(* the following should really be made generic *)
(* on union maps, not just histories *)
(* moreover, it's decidable for maps with decidable contents equality *)
(* but I don't want to bother now *)
(* the valid h1 = valid h2 condition is imposed so that we can get *)
(* extensionality: i.e., h1 <<= h2 and h1(x) = v then h2(x) = v *)
(* this property only holds if h2 is valid too, thus we rule out *)
(* that invalid elements can be compared to valid ones *)
Definition hist_preord S (h1 h2 : hist S) := 
  valid h1 = valid h2 /\ exists h, h2 = h1 \+ h.

Notation "'[' h1 '<<=' h2 ']'" := (hist_preord h1 h2).

(* Do we really need the next notation? *)
(* Notation "'[' h '<<<' t ']'"   := (last_key h <= t).  *)

Lemma hist_refl S (h : hist S) : [h <<= h].
Proof. by split=>//; exists Unit; rewrite unitR. Qed.

Hint Resolve hist_refl.

Lemma hist_trans S (h1 h2 h3 : hist S) : [h1 <<= h2] -> [h2 <<= h3] -> [h1 <<= h3].
Proof. 
case=>E1 [h1' F1][E2][h2' F2]; split; first by rewrite E1 E2.
by rewrite F2 F1 -joinA; exists (h1' \+ h2').
Qed.

Lemma hist_asym S (h1 h2 : hist S) : [h1 <<= h2] -> [h2 <<= h1] -> h1 = h2.
Proof.
case=>_ [h1' ->{h2}][_][h2']; case V: (valid h1); last first.
- by move/invalidE: (negbT V)=>->; rewrite /PCM.join /= !umE.
rewrite -{1 2}[h1]unitR -joinA in V *. 
by case/(joinfK V)/esym/join0E=>->; rewrite unitR. 
Qed.

Lemma hist_unit S (h: hist S) : valid h -> [Unit <<= h].
Proof. 
rewrite /hist_preord valid_unit. 
by split=>//; exists h; rewrite unitL. 
Qed.

(* the other implication only holds if A is non-empty *)
(* so I can't prove it in general *)
Lemma histE S (h1 h2: hist S) s t : 
        [h1 <<= h2] -> [s in h1 at t] -> [s in h2 at t].
Proof.
case=>E [h] F [s' E']; exists s'; rewrite {h2}F in E *.
rewrite findUnL ?(find_some E') // -E.
by apply: dom_valid (find_some E').
Qed.

Lemma histK S (h1 h2 : hist S) : [h1 <<= h2] -> last_key h1 <= last_key h2.
Proof.
case=>V /= [h] E; rewrite {h2}E in V *. 
apply: last_key_mono=>/= x D.
by rewrite domUn !inE /= D -V (dom_valid D). 
Qed.

Lemma hist_plus_mono S (h1 h2 h : hist S) :
  valid(h2 \+ h) -> [h1 <<= h2] -> [h1 \+ h <<= h2 \+ h].
Proof.
move=>V2 H; split; case: H=> E[h']Z; subst h2.
by rewrite V2; rewrite joinAC in V2; move/validL: V2. 
by exists h'; rewrite joinAC.
Qed.

(* Aleks: do we really need this? it's a simple transitivity *)
(* from last_key h1 <= last_key (h1 \+ h2), which is last_key_mono *)
Lemma last_key_ltL S (h1 h2: hist S) t :
   valid (h1 \+ h2) -> last_key (h1 \+ h2) < t -> last_key h1 < t.
Proof.
move=>V; apply: leq_trans; apply: last_key_mono.
by move=>x D; rewrite domUn inE V D. 
Qed.

(* All these are not even properties of histories, but of unionmaps *)
(* should be moved to union maps *)
(* come up with better names *)
Lemma hist_plus_mono2 S (h1 h2 h : hist S) :
        valid (h1 \+ h) -> [h1 <<= h2 \+ h] -> [h1 <<= h2].
Proof.
suff L t (u : S * S) h1' h2' : valid (h1' \+ t \\-> u) -> 
  [h1' <<= h2' \+ t \\-> u] -> [h1' <<= h2'].
- elim/gen_ind': h h1 h2 => [||k v f IH V] h1 h2.
  - by move/validR; rewrite -[valid _]negbK; move/negP/invalidE.
  - by rewrite !unitR. 
  rewrite !joinA joinAC => H1 /(IH _ _ (validL H1)).
  by rewrite joinAC in H1; apply: L (validL H1).
move=>V' [V][h' E].
have V'' : valid (h2' \+ t \\-> u) by rewrite -V (validL V').
split; first by rewrite V V'' (validL V'').
exists (free t h'); move: (f_equal (fun x => free t x) E).
rewrite joinC in V'' *; rewrite um_freePtUn // freeUnL //. 
by rewrite joinC in V'; apply: um_validPtD V'.
Qed.

Lemma last_key_less S (h: hist S) t u:
  valid h -> last_key h < t -> valid (h \+ t \\-> u).
Proof.
move=> V1 L1; rewrite joinC um_validPtUn; apply/andP; split=>//. 
apply:(contra (b:=false))=>//; move/last_key_max.
rewrite ltnNge in L1; move: L1. 
by move/negbTE=>->. 
Qed.

(* come up with better names *)
Lemma joinhK S (h1 h2 h : hist S) :
        valid (h \+ h1) -> [h \+ h1 <<= h \+ h2] -> [h1 <<= h2].
Proof.
move=>V' [/esym V][h'] E; rewrite V' in V.
split; first by rewrite (validR V) (validR V'). 
by exists h'; rewrite -joinA in E; rewrite (joinfK V E). 
Qed.

Lemma hist_monoL S (h1 h2 h : hist S) : 
        valid (h \+ h2) -> [h1 <<= h2] -> [h \+ h1 <<= h \+ h2].
Proof. by rewrite !(joinC h); apply: hist_plus_mono. Qed.

(* backward induction on histories *)
Lemma hist_indb S (P : hist S -> Prop) : 
        P Unit ->
        (forall u, P (0 \\-> u)) ->
        (forall t u f, P f -> last_key f < t -> 
                       valid (f \+ t \\-> u) -> P (f \+ t \\-> u)) ->
        forall f, valid f -> P f.
Proof.
move=>H1 H2 H3; elim/gen_indb=>//=.
- by rewrite -[valid _]negbK; move/negP/invalidE.
move=>k v f H V K _; rewrite joinC in V *.
case E: (empb f); last first.
- apply: H3=>//; first by apply: H (validL V).
  apply: K; apply: contraR (negbT E).
  by rewrite keys_dom -empbE; apply: last_key_dom (validL V).
move/empbE: E=>->; rewrite unitL.
case: k K V=>[|k] K V; first by apply: H2.
move: (H3 k.+1 v Unit H1).  
by rewrite last_key0 unitL; apply=>//; apply: um_validPt.
Qed.

(* forward induction on histories *)
Lemma hist_indf S (P : hist S -> Prop) : 
        P Unit ->
        (forall t u f, P f -> 
           (forall x, x \in dom f -> t < x) ->
           valid (f \+ t \\-> u) -> P (f \+ t \\-> u)) ->
        forall f, valid f -> P f.
Proof.
move=>H1 H2; elim/gen_indf=>//=.
- by rewrite -[valid _]negbK; move/negP/invalidE.
move=>k v f H V K _; rewrite joinC in V *.
apply: H2=>//; first by apply: H (validL V).
move=>x; move/(order_path_min (@ordtype.trans _)): K. 
by case: allP=>// X _; rewrite -keys_dom; apply: X. 
Qed.

Lemma hist_cancelV S (h1 h2 : hist S) t1 t2 u1 u2 : 
        valid (t1 \\-> u1 \+ h1) -> 
        t1 \\-> u1 \+ h1 = t2 \\-> u2 \+ h2 ->
        if t1 == t2 then [/\ u1 = u2 & h1 = h2]
        else [/\ h1 = t2 \\-> u2 \+ free t1 h2 & 
                 h2 = t1 \\-> u1 \+ free t2 h1].
Proof.
case: ifP=>E V X.
- by move/eqP: E X=><- /(um_cancel V) [].
rewrite -{1}(um_freePtUn V) X; rewrite X in V.
rewrite -{3}(um_freePtUn V) -{2}X.
by rewrite !freeUnL // um_domPt inE ?E // eq_sym E.
Qed.

Lemma cancelPtUn S t1 t2 u1 u2 (h: hist S):
   valid (t1 \\-> u1) ->  
   t1 \\-> u1 = (t2 \\-> u2 \+ h) ->
   [/\ t1 = t2, u1 = u2 & h = Unit].
Proof.
move=> V E.
have: find t2 (t2 \\-> u2 \+ h) = find t2 (t1 \\-> u1) by rewrite -E.
rewrite !gen_findPtUn; last by rewrite E in V.
rewrite sym; move/find_some; rewrite um_domPt inE; move/eqP=>Z; subst t2.
rewrite -(unitR (t1 \\-> _)) in V E.
by case/(gen_cancel V): E=>-> _ <-. 
Qed.

Lemma complete_shape S (h: hist S) t s1 s2:
  0 < t -> complete (h \+ t \\-> (s1, s2)) -> 
  [s1 in h \+ t \\-> (s1, s2) at t.-1].
Proof.
move=>N; rewrite joinC; case=>Wf; move/wf_valid: (Wf)=>V.
have Z: t \in dom (t \\-> (s1, s2) \+ h).
- by move:(um_findPtUn V)=>/find_some.
have X: t.-1 <= last_key (t \\-> (s1, s2) \+ h).
suff: t <= last_key (t \\-> (s1, s2) \+ h); first by apply:leq_trans (leq_pred t).
- by move/last_key_max: Z.
move/(_ (t.-1) X)=> D1.
case:(Wf)=>_; move/(_ t.-1 D1); rewrite (prednK N); case=>//.
move=>z[z1][z2][F1]F2; move:(um_findPtUn V); rewrite F2=>{F2}. 
by case=>Z1 Z2; subst z1 z2; exists z.
Qed.

Lemma complete_not_unit S (h : hist S) :
  complete h -> h = Unit -> False.
Proof.
move=>C Z; subst h.
by case:C=>_ /(_ 0 (leqnn 0)); rewrite dom0 inE.
Qed.

Lemma complete_not_unit' S (h : hist S) u:
  complete (h \+ fresh h \\-> u) -> h = Unit -> False.
Proof.
rewrite joinC; move=>C; case:(C)=>[[H1 H2] H].
case: H1=>s; rewrite (findUnL 0 (complete_valid C)); case:ifP=>[|_].
- rewrite /fresh um_domPt inE =>/eqP Z. 
  by move:(ltn0Sn (last_key h)); rewrite Z; rewrite (ltnn 0).
  by move/find_some=>G1 G2; subst h; rewrite dom0 inE in G1.
Qed.

Lemma cont_shape S (h : hist S) P t s1 s2:
  0 < t -> complete (h \+ t \\-> (s1, s2)) -> 
  cont P (h \+ t \\-> (s1, s2)) -> P s1 s2.
Proof.
move=> N C Sc; case:(complete_shape N C)=>z F1; rewrite joinC in C Sc F1 *.
move/complete_valid: (C)=>V; move:(um_findPtUn V)=>F2.
move:(Sc t.-1 s1 s2); rewrite (prednK N). 
have S1: [s1 in t \\-> (s1, s2) \+ h at t.-1] by exists z. 
have S2: [s2 in t \\-> (s1, s2) \+ h at t] by exists s1.
by move/(_ S1 S2).
Qed.

Lemma complete_last_key S (h : hist S) u: 
  complete (h \+ fresh h \\-> u) -> 
  last_key h \in dom (h \+ fresh h \\-> u) /\
  fresh h \in dom (h \+ fresh h \\-> u).
Proof.
move=>C; move:(complete_valid C)=>V; split.
- rewrite domUn inE; apply/andP; split; first by move:(complete_valid C).
  apply/orP; left.
  case X: (last_key h \in dom h)=>//.
  move:(last_key_dom (validL V)); rewrite X=>/=; move/(_ is_true_true).
  by move/(complete_not_unit' C).
rewrite joinC um_domPtUn inE. apply/andP; split=>//; last by apply/orP; left.
by rewrite joinC; move:(complete_valid C).
Qed.

Lemma complete_last_fresh S (h : hist S) t u:
  complete (h \+ t \\-> u) -> last_key h < t ->
  t = fresh h.
Proof.
rewrite joinC; move=>C N; move/complete_valid: (C)=> V.
have Dt: t \in dom (t \\-> u \+ h) by move/find_some: (um_findPtUn V).
case X: (t == (last_key h).+1); first by move/eqP: X.
case:(C)=>[_]H; move:(H (fresh h)).
have Y: last_key (t \\-> u \+ h) = t.
- apply: max_key_last=>// x /=.
  rewrite domUn inE; case/andP=>_; case/orP.
  - by rewrite um_domPt inE /=; move/eqP=>->; apply:leqnn.
  by move/last_key_max; move/leq_ltn_trans /(_ N) /ltnW.
rewrite Y leq_eqVlt eq_sym X /=.
have Z: fresh h < t.
rewrite leq_eqVlt in N; case/orP: N =>//; by rewrite eq_sym X.
move/(_ Z); rewrite domUn inE /=; case/andP=>[_]; case/orP.
- by rewrite um_domPt inE /=; move/eqP.
by move/dom_ordfresh; rewrite ltnn.
Qed.

Lemma  wf_fresh S (h : hist S)u: 
  well_formed (h \+ fresh h \\-> u) -> well_formed h.
Proof.
move=>Wf; move/wf_valid: (Wf)=> V.
case: (Wf)=>[H1]H2; split=>[|k G1 G2].
- case: H1=>s; rewrite findUnL //. 
  case: ifP=>[_|_]; first by move=>G; exists s.
  move/find_some. rewrite um_domPt inE eq_sym /fresh /=.
  by move/eqP=>E; move:(ltnn 0); rewrite {2}E ltn0Sn.
move:(H2 k).
case; do?[by rewrite domUn inE; apply/andP; split=>//; apply/orP; left]. 
move=>s[s1][s2].
case K: (k == fresh h).
- by move/eqP: K=>K; subst k; move/dom_ordfresh: G1; rewrite ltnn.
case K1: (k.+1 == fresh h).
- by move/eqP: K1=>K1; rewrite K1 in G2; move/dom_ordfresh: G2; rewrite ltnn.
rewrite !findUnL=>//; case:ifP=>[_[Y1]|_[]].
- case:ifP=>[_ Y2|_]; first by exists s, s1, s2. 
  by move/find_some; rewrite um_domPt inE eq_sym K1.
by move/find_some; rewrite um_domPt inE eq_sym K.
Qed.

Lemma cont_fresh S (h: hist S) P u :
  valid (h \+ fresh h \\-> u) -> cont P (h \+ fresh h \\-> u) -> cont P h.
Proof.
move=>V Sc.
move=>t' s1' s2' H1 H2; move:(Sc t' s1' s2')=>Hs; apply: Hs.
- move/last_key_in: (H1)=>L1'; move:(leq_ltn_trans L1' (ltnSn (last_key h)))=>N1.
  case: H1=>t1 Hf1; exists t1; rewrite findUnR //.
  by rewrite um_domPt !inE /=; move/gtn_eqF: N1=>->.
move/last_key_in: (H2)=>L2'; move:(leq_ltn_trans L2' (ltnSn (last_key h)))=>N2.
case: H2=>t2 Hf2; exists t2; rewrite findUnR //.
by rewrite um_domPt !inE /=; move/gtn_eqF: N2=>->.
Qed.

Lemma complete_fresh S (h : hist S) u :
  complete (h \+ fresh h \\-> u) -> complete h.
Proof.
case=>[Wf]H; split; first by apply: wf_fresh Wf.
move=>k; move:(H k); case X: (k == fresh h).
- by move=>_; move/eqP: X=>->; rewrite leqNgt ltnSn. 
rewrite last_fresh; last by move/validL: (wf_valid Wf). 
rewrite leq_eqVlt X /= ltnS=>H1; move/H1.
rewrite domUn !inE /=; case/andP=>_; case/orP=>//.
by rewrite um_domPt inE /= eq_sym X. 
Qed.

Lemma complete0 (S: encoded_set) (s: S): complete (0 \\-> (s, s) : hist S).
Proof.
split=>[|k]; last first. 
- by rewrite last_keyPt leqn0; move/eqP=>->{k}; rewrite um_domPt inE /=. 
split=>[|k _]; first by exists s; rewrite um_findPt.
by rewrite um_domPt inE /=; move/eqP=>Z; move:(ltn0Sn k); rewrite Z ltnn. 
Qed.

Lemma contPt (S: encoded_set) P t (s1 s2: S): cont P (t \\-> (s1, s2) : hist S).
Proof.
move=>t1 t2 s [z1]/find_some; rewrite um_domPt inE /=; move/eqP=><-{t1}.
move=>[z2]/find_some; rewrite um_domPt inE /=; move/eqP=>Z.
by move:(ltnSn t); rewrite -{2}Z ltnn.
Qed.
    
