Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path.
Require Import pred prelude idynamic ordtype finmap pcm coding unionmap heap.
Require Import feap state zigzag world inject rely hide atomic autosplit.
Require Import privstate footprint model lemmas locks stsep getter.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(* An FCSL Hello World example - concurrent incrementor *)

Module Incrementor.
Section Incrementor.
Variables (hp lk : nat) (x : ptr).

Definition incr_inv (g : nat) := [Pred h | h = x :-> g].

Lemma incr_inv_prec (g : [encoded_pcm of nat]) : precise (incr_inv g).
Proof. by move=>h1 h2 t1 t2 V1 E; move=>->->. Qed.

Variable IncrM : lockmod (LockInfo lk incr_inv_prec). 
Notation W := (entangle (Priv hp) IncrM). 
Notation is_lock := (is_lock IncrM).
Notation locked := (locked IncrM).
Notation lock := (lock IncrM hp).
Notation unlock := (unlock IncrM hp).
Notation coh := (Mod.coh W).
Notation pv_getso s := (getso (pvg hp) s).
Notation pv_self s := (self (pvg hp) s).
Notation pv_other s := (other (pvg hp) s).
Notation "[ e ]" := (extend _ e).

Lemma pfF n : @local [encoded_pcm of nat] (addn^~ n). 
Proof. by move=>m y; rewrite /= -addnAC. Qed.

(* Coarse-grained incrementor procedure *)

Program Definition incr n : 
 {gs}, STsep [W] 
         (fun i => exists sl go,
            i = hp ->> [Unit, tt, pv_other i] \+ sl /\ is_lock gs go sl,
          fun (_: unit) m => exists sl go,
            m = hp ->> [Unit, tt, pv_other m] \+ sl /\ 
            is_lock (gs + n) go sl) :=  
  Do (lock;;
      tmp <-- [priv_read hp nat x];
      [priv_write hp x (tmp + n)];;
      unlock (@pfF n)).
Next Obligation. by apply: entangle_extL. Defined.
Next Obligation. by apply: entangle_extL. Defined.
Next Obligation.
apply: gh=>i1 gs [sl][go][Ei1] Isl1 Ci1; rewrite Ei1 in Ci1 *=>{Ei1}.
apply: step; apply: val_do'=>[|[] i2].
- by exists Unit, (pv_other i1), gs, go, sl; split=>//.
case/(_ Unit (pv_other i1) gs go sl erefli Isl1)=>h[ge'][he'][sl2][Ei2]Lk V Inv Ci2.
rewrite /= in Inv; subst h i2; rewrite unitL in Ci2 *.
apply: step; apply val_extend.  
- move/cohH: (Ci2). rewrite /validH flatsUn; last by move/cohV: Ci2.
  rewrite priv_flatsE=>/validL=>V1; split; first by rewrite /validH priv_flatsE.
  by exists (x :-> (gs + ge')), he'.
apply: (gh_ex (gs + ge')); apply: (gh_ex Unit). 
apply:val_do=>[_|m sp3[->{m}]]; first by rewrite unitR getE.
rewrite unitR=> Ep3 Cp3 sl3 Ci3 Mp3.
apply:step; apply: val_extend=>//; apply:(gh_ex Unit). 
apply:val_do=>[_|_ sp4]; first by exists nat, (gs + ge'); rewrite Ep3 unitR.
rewrite unitR=>Ep4 Cp4 sl4 Ci4 Mp4.
move:(locked_stable Lk (menvs_trans Mp3 Mp4))=>Lk4.
apply: val_do'=>/=[|[]i5].
- exists gs, ge', Unit, (pv_other sp4), (x :-> (gs + ge' + n)), sl4; split=>//.
  - by rewrite /PCM.join /= -2!(addnC n) addnA.
  congr (_\+_); rewrite unitL. 
  by move/cohE:(Cp4)=>/=; rewrite Ep4; case: Cp4=>_[?][?]=>->; rewrite getE.
case/(_ gs ge' Unit (pv_other sp4) (x :-> (gs + ge' + n)) sl4). 
- split=>//; first by rewrite /= -2!(addnC n) addnA.
  congr (_\+_); rewrite unitL. 
  by move/cohE:(Cp4)=>/=; rewrite Ep4; case: Cp4=>_[?][?]=>->; rewrite getE.
move=>he5[ge5][sl5][Ei5]Isl5 Ci5; exists sl5, ge5; split=>//.
by subst i5; rewrite getE=>//; move/cohV: Ci5. 
Qed.

End Incrementor.

End Incrementor.
