Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path.
Require Import pred prelude idynamic ordtype finmap pcm coding unionmap heap.
Require Import feap state zigzag world inject rely hide atomic autosplit finfun.
Require Import pperm privstate footprint model lemmas stsep histories treiber.
Require Import stack_framed array fintype tuple getter alloc stack_seq.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(*
1. TODO: generalize getters so that a getter for star_coh is a union
of individual getters.
*)

(************************************)
(* first, stuff about linked lists  *)
(* this should be a generic library *)
(************************************)
Section ClientPredicates.
Variable A: encoded_set.
Notation st := [encoded_set of seq A].
Definition hist := hist [encoded_set of (seq A)].
Definition P_stack s1 s2 := exists e : A, s1 = e::s2 \/ s2 = e::s1.
Definition stacks_cont := cont P_stack.

Definition all_pushed2 (h : hist) := 
  forall t s1 s2, find t h = Some (s1, s2) ->
  [/\ t = 0 & s1 = s2] \/ exists v, s2 = v :: s1.


Fixpoint popped (h : hist) (xs : seq A) : Prop := 
  if xs is x :: xs' then 
    exists t s h', h = t \\-> (x :: s, s) \+ h' /\ 
                     popped h' xs'
  else all_pushed2 h. 

Definition all_popped2 (h : hist) := 
  forall t s1 s2, find t h = Some (s1, s2) -> 
  [/\ t = 0, s1 = s2 & s2 = [::]] \/
  exists v, s1 = v :: s2.

Fixpoint pushed (h : hist) (xs : seq A) : Prop := 
  if xs is x :: xs' then 
    exists t s h', [/\ (* last_key h' < t, *)
                        h = t \\-> (s, x::s) \+ h' &
                        pushed h' xs']
  else all_popped2 h. 


(* What is currently in the stack *)
Definition in_stack (h: hist) (xs: seq A) : Prop :=
  h = Unit /\ xs = [::] \/
  [xs in h at last_key h].  

(* Lemmas *)

Lemma pushed_mono h1 h2 (L : seq A) : 
        pushed h1 L -> all_popped2 h2 -> pushed (h1 \+ h2) L.
Proof.
elim: L h1 h2=>[|a L IH] h1 h2 /=; last first.
- case=>t [s][h'][H] P /(IH _ _ P) P'. 
  by exists t, s, (h' \+ h2); rewrite H joinA. 
move=>H1 H2 t s1 s2 F. 
move/dom_valid: (find_some F) (find_some F)=>V.
rewrite domUn inE V /=; case/orP=>D; move: F. 
- by rewrite findUnL // D; apply: H1.
by rewrite findUnR // D; apply: H2.
Qed.

Lemma popped_mono h1 h2 (L : seq A) :  
        popped h1 L -> all_pushed2 h2 -> popped (h1 \+ h2) L.
Proof.
elim: L h1 h2=>[|a L IH] h1 h2 /=; last first.
- case=>t [s][h'][H] P /(IH _ _ P) P'.
  by exists t, s, (h' \+ h2); rewrite H joinA. 
move=>H1 H2 t s1 s2 F.
move/dom_valid: (find_some F) (find_some F)=>V. 
rewrite domUn inE V /=; case/orP=>D; move: F.
- by rewrite findUnL // D; apply: H1. 
by rewrite findUnR // D; apply: H2.
Qed.

Lemma popped_perm h (xs xs' : seq A) : 
        popped h xs -> perm xs xs' -> popped h xs'.
Proof.
move=>H1 H2; elim: xs xs' / H2 h H1 => xs xs'; last 1 first.
- by move=>t H1 IH1 H2 IH2 h /IH1/IH2. 
- by move=>->->. 
- move=>x t1 t2 ->-> H IH h /=.
  by case=>t [s][h'][->] H1; exists t, s, h'; split=>//; apply: IH H1.
move=>x y t ->-> h /=. 
case=>t1 [s1][h1][->][t2][s2][h2][->] H2. 
exists t2, s2, (t1 \\-> (x :: s1, s1) \+ h2).
by rewrite joinCA; split=>//; exists t1, s1, h2. 
Qed.

(* This should be unified w/ the previous lemma! *)
Lemma pushed_perm h (xs xs' : seq A) : 
        pushed h xs -> perm xs xs' -> pushed h xs'.
Proof.
move=>H1 H2; elim: xs xs' / H2 h H1 => xs xs'; last 1 first.
- by move=>t H1 IH1 H2 IH2 h /IH1/IH2. 
- by move=>->->. 
- move=>x t1 t2 ->-> H IH h /=.
  by case=>t [s][h'][->] H1; exists t, s, h'; split=>//; apply: IH H1.
move=>x y t ->-> h /=. 
case=>t1 [s1][h1][->][t2][s2][h2][->] H2. 
exists t2, s2, (t1 \\-> (s1, x :: s1) \+ h2).
by rewrite joinCA; split=>//; exists t1, s1, h2. 
Qed.

Lemma inv_push_unit xs : pushed Unit xs -> xs = [::].
Proof.
case: xs=>//= x xs [t][s][h][] /esym /empbP.
by rewrite empbUn um_ptsU empbU.
Qed.

Lemma inv_pop_unit xs : popped Unit xs -> xs = [::].
Proof.
case: xs=>//= x xs [t][s][h][] /esym /empbP.
by rewrite empbUn um_ptsU empbU.
Qed.

(* silly lemmas, but they appear all over the place *)
Lemma seq_contra1 (x1 : A) xs : [:: x1 & xs] = xs -> False.
Proof. by elim: xs x1=>[|x xs IH] x1 // [<-] /IH. Qed.

Lemma seq_contra2 (x1 x2 : A) xs : [:: x1, x2 & xs] = xs -> False.
Proof. by elim: xs x1 x2=>[|x xs IH] x1 x2 // [<-] /IH. Qed.

(* inversion for push *)
Lemma push_good_find t xs s v h : 
        valid h ->  
        find t h = Some (s, v :: s) ->
        pushed h xs ->
        exists xs', perm xs (v :: xs') /\ pushed (free t h) xs'.
Proof.
elim: xs t s v h=>[|x xs IH] t s v h V E /=.
- by case/(_ _ _ _ E); [case=>_ -> | case=>w /esym /seq_contra2].
case=>t1 [s1][h1][] E'; rewrite {h}E' in V E *.
case X : (t == t1) E.
- rewrite (eqP X) um_findPtUn // um_freePtUn //. 
  by case=><-<-; exists xs. 
rewrite findUnL // freeUnL um_domPt inE eq_sym X //. 
move/validR: V=>V F /(IH _ _ _ _ V F) [xs'][P H].
exists (x :: xs'); split; last by exists t1, s1, (free t h1). 
by apply: (@perm_cons_cat_consL _ [:: v]) P.
Qed.

Lemma push_bad_find t xs s v h : 
        valid h ->  
        find t h = Some (v :: s, s) ->
        pushed h xs ->
        pushed (free t h) xs.
Proof.
elim: xs t s v h=>[|x xs IH] t s v h V E /=.
- move=>H t' s1 s2. 
  case X : (t == t'); first by rewrite -{t' X}(eqP X) findF eq_refl.
  by rewrite findF eq_sym X; apply: H. 
case=>t1 [s1][h1][] E'; rewrite {h}E' in V E *.
case X : (t == t1) E.
- by rewrite (eqP X) um_findPtUn // um_freePtUn //; case=>-> /seq_contra2. 
rewrite findUnL // freeUnL um_domPt inE eq_sym X //. 
move/validR: V=>V F /(IH _ _ _ _ V F) H.
by exists t1, s1, (free t h1).
Qed.

Lemma inv_push_good_cons t xs s v h : 
        valid (h \+ t \\-> (s, v :: s)) ->  
        pushed (h \+ t \\-> (s, v :: s)) xs ->
        exists xs', perm xs (v :: xs') /\ pushed h xs'.
Proof.
rewrite joinC => V. 
have X : find t (t \\-> (s, v::s) \+ h) = Some (s, v::s).
- by rewrite um_findPtUn.
case/(push_good_find V X)=>xs' [P H].
by exists xs'; rewrite um_freePtUn // in H.
Qed.

Lemma inv_push_bad_cons t xs s v h : 
        valid (h \+ t \\-> (v :: s, s)) ->  
        pushed (h \+ t \\-> (v :: s, s)) xs ->
        pushed h xs.
Proof.
rewrite joinC => V.
have X : find t (t \\-> (v :: s, s) \+ h) = Some (v :: s, s).
- by rewrite um_findPtUn.
by move/(push_bad_find V X); rewrite um_freePtUn.
Qed.


(* inversion for pop *)
Lemma pop_good_find t xs s v h : 
        valid h ->  
        find t h = Some (v :: s, s) ->
        popped h xs ->
        exists xs', perm xs (v :: xs') /\ popped (free t h) xs'.
Proof.
elim: xs t s v h=>[|x xs IH] t s v h V E /=.
- by case/(_ _ _ _ E); [case=>_ /seq_contra1 | case=>x /esym /seq_contra2]. 
case=>t1 [s1][h1][] E'; rewrite {h}E' in V E *.
case X : (t == t1) E.
- rewrite (eqP X) um_findPtUn // um_freePtUn //. 
  by case=><-<-; exists xs. 
rewrite findUnL // freeUnL um_domPt inE eq_sym X //. 
move/validR: V=>V F /(IH _ _ _ _ V F) [xs'][P H].
exists (x :: xs'); split; last by exists t1, s1, (free t h1). 
by apply: (@perm_cons_cat_consL _ [:: v]) P.
Qed.

Lemma pop_bad_find t xs s v h : 
        valid h ->  
        find t h = Some (s, v :: s) ->
        popped h xs ->
        popped (free t h) xs.
Proof.
elim: xs t s v h=>[|x xs IH] t s v h V E /=.
- move=>H t' s1 s2. 
  case X : (t == t'); first by rewrite -{t' X}(eqP X) findF eq_refl.
  by rewrite findF eq_sym X; apply: H. 
case=>t1 [s1][h1][] E'; rewrite {h}E' in V E *.
case X : (t == t1) E.
- by rewrite (eqP X) um_findPtUn // um_freePtUn //; case=><- /esym /seq_contra2. 
rewrite findUnL // freeUnL um_domPt inE eq_sym X //. 
move/validR: V=>V F /(IH _ _ _ _ V F) H.
by exists t1, s1, (free t h1).
Qed.

Lemma inv_pop_good_cons t xs s v h : 
        valid (h \+ t \\-> (v :: s, s)) ->  
        popped (h \+ t \\-> (v :: s, s)) xs ->
        exists xs', perm xs (v :: xs') /\ popped h xs'.
Proof.
rewrite joinC => V. 
have X : find t (t \\-> (v :: s, s) \+ h) = Some (v :: s, s).
- by rewrite um_findPtUn.
case/(pop_good_find V X)=>xs' [P H].
by exists xs'; rewrite um_freePtUn // in H.
Qed.

Lemma inv_pop_bad_cons t xs s v h : 
        valid (h \+ t \\-> (s, v :: s)) ->  
        popped (h \+ t \\-> (s, v :: s)) xs ->
        popped h xs.
Proof.
rewrite joinC => V.
have X : find t (t \\-> (s, v :: s) \+ h) = Some (s, v :: s).
- by rewrite um_findPtUn.
by move/(pop_bad_find V X); rewrite um_freePtUn.
Qed.

(* Two useful in_stack lemmas *)
Lemma in_stack_pushed (h: hist) v s L:
  complete (h \+ fresh h \\-> (s, v :: s)) ->
  in_stack h L -> L = s.
Proof.
move=>C; case.
- move=>[Z1] Z2; subst h L; rewrite unitL /fresh last_key0 in C.
  by case:(C)=>Wf _; case:Wf=>[[s']]; move/find_some; rewrite um_domPt inE=>/eqP. 
move:(complete_valid C)=> V.
case=>z=>F1; rewrite joinC in V; move:(um_findPtUn V)=> F2. 
rewrite /fresh in F1 F2. case:(C)=>[[_] H] _. 
case:(complete_last_key C)=> G1 G2; case:(H (last_key h) G1 G2)=>z'[z1][z2][E1]E2.
rewrite joinC F2 in E2; case: E2=>Z1 Z2; subst z1 z2.
rewrite (findUnL (last_key h) (complete_valid C)) in E1. 
case:ifP E1=>_; first by rewrite F1; case. 
move/find_some; rewrite um_domPt inE /fresh=>/eqP=> Z.
by move:(ltnn (last_key h)); rewrite -{2}Z (ltnSn (last_key h)).
Qed.

Lemma in_stack_popped (h: hist) v s L:
  complete (h \+ fresh h \\-> (v::s,s)) ->
  in_stack h L -> L = v::s.
Proof.
move=>C; case.
- move=>[Z1] Z2; subst h L; rewrite unitL /fresh last_key0 in C.
  by case:(C)=>Wf _; case:Wf=>[[s']]; move/find_some; rewrite um_domPt inE=>/eqP. 
case:(complete_last_key C)=> G1 G2. 
move:(complete_valid C)=> V.
case:(C)=>Wf H; case:Wf=>[H1] /(_ (last_key h) G1 G2); case.
move=>z[z1][z2][E1]; rewrite joinC in V *; move:(um_findPtUn V)=>->.
case=>Z1 Z2; subst z1 z2. 
rewrite (findUnL (last_key h) (complete_valid C)) in E1. 
case:ifP E1=>_; first by move=>E1[z']; rewrite E1; case=>_->.
move/find_some; rewrite um_domPt inE /fresh=>/eqP=> Z.
by move:(ltnn (last_key h)); rewrite -{2}Z (ltnSn (last_key h)).
Qed.

(* now the main lemma *)
Lemma push_pop_all h (L1 L2 : seq A) :
        complete h -> stacks_cont h -> pushed h L1 -> popped h L2 -> 
        exists (L: seq A), in_stack h L /\ perm L1 (L2 ++ L).
Proof.
move=>C; move: (wf_valid (proj1 C)) => Vh; move: h Vh L1 L2 C. 
apply: hist_indb=>[|[s' s]|t [s1 s2] h IH K V] L1 L2 C Sc Hpu Hpo. 
- move/inv_push_unit: Hpu Hpo=>-> /inv_pop_unit ->. 
  by exists [::]; split=>//; left. 
- case: (C); case=>[[z]]Ez _ _; rewrite um_findPt in Ez;case:Ez=>Z1 Z2;subst s' s.
  have X: L2 = [::].
  case: L2 Hpo=>//= a ls [t][s1][h']; case. 
  by case/(cancelPtUn (complete_valid C))=>_;case=>->; move/seq_contra1.
  have Y: L1 = z.
  case: L1 Hpu=>//=[|a l [t][s][h'][]]. 
  + move/(_ 0 z z); rewrite um_findPt; case=>//; case=>//x;
    last by rewrite sym; move/seq_contra1. 
  + by case/(cancelPtUn (complete_valid C))=>_; case=>->; rewrite sym; move/seq_contra1.
  subst z L2. exists L1. 
  split=>//; by right; rewrite last_keyPt um_findPt; exists L1.  
move:(complete_last_fresh C K)=>Z; subst t.
have C1: complete h by apply:(complete_fresh C).
have Sc1: stacks_cont h by apply: cont_fresh (complete_valid C) Sc.
have [s [v X]] : exists s v, (s1, s2) = (s, v :: s) \/ (s1, s2) = (v :: s, s).
have K1 : 0 < fresh h by apply: (ltn0Sn (last_key h)) .
move: (cont_shape (P:=P_stack) (t := (fresh h)) (h:=h))=>H.
move:(H s1 s2 K1 C Sc)=>[v]. 
case=>->;[by exists  s2, v; right| by exists s1, v; left].
case: X=>X; case: X=>Z1 Z2; subst s1 s2.
- case/(inv_push_good_cons V): Hpu=>L1' [P Hpu].
  move/(inv_pop_bad_cons V): Hpo=>Hpo.
  case: (IH _ _ C1 Sc1 Hpu Hpo)=>L2' [In P'].
  have Y : L2' = s by apply: in_stack_pushed C In.
  subst L2'.
  exists (v :: s); split; last first.
  - apply: perm_trans P _.
    by apply/perm_cons_cat_cons.
  right; move/validL: V => V1; rewrite (last_fresh (s,v::s) V1).
  exists s; move/complete_valid: C; rewrite joinC=>V.
  by apply:(um_findPtUn V).
move/(inv_push_bad_cons V): Hpu=>Hpu. 
case/(inv_pop_good_cons V): Hpo=>L2' [P Hpo]. 
case: (IH _ _ C1 Sc1 Hpu Hpo)=>L1' [In P'].
have Y : L1' = v :: s by apply:in_stack_popped C In.
subst L1'.
exists s; split; last first.
- apply: perm_trans P' _; apply/perm_sym.  
  move/(perm_cat2r s): P=>P. 
  by apply: perm_trans P (perm_cons_catCA _ _ v).
right; rewrite last_fresh; last by move/validL: V.
exists (v::s); move:(findUnL (fresh h) V)=>->.
case: ifP=>[H|_]; last by rewrite um_findPt.
move: (dom_fresh h); move/contra. move/(_ true). 
have X: true -> fresh h \in dom h by [].
by move/(_ X). 
Qed.

End ClientPredicates.



Section ClientPrograms.
Variables (pv: nat) (T : treiberT).
Notation A := (tb_set T). 
Notation sent := (tb_sent T).
Notation l := (tb_label T).
Notation st := [encoded_set of seq A].
Notation W := (entangle (Priv pv) (Treiber T)). 
Notation pt_coh := (Mod.coh W).
Notation pv_coh := (Mod.coh (Priv pv)).
Notation tb_coh := (Mod.coh (Treiber T)).
Notation guar := (Mod.internal W). 
Notation tb_getso s := (getso (tbg T) s).
Notation tb_self s := (self (tbg T) s).
Notation tb_other s := (other (tbg T) s).
Notation tb_joint s := (joint (tbg T) s).
Notation pv_getso s := (getso (pvg pv) s).
Notation pv_self s := (self (pvg pv) s).
Notation pv_other s := (other (pvg pv) s).
Variable n : nat.
(* Variable Ac : {array [finType of 'I_n] -> A}. *)

Definition is_array p := @shape [finType of 'I_n] A (Array p).

Notation "[ e ]" := (extend _ e).

Lemma take_plus1 B i (s : seq B) (v : B) : 
        i < size s -> take i.+1 s = take i s ++ [:: nth v s i].
Proof.
move: take_nth => L; move=>N; move: (L _ v _ _ N)=>->.
by rewrite cats1.
Qed.

Lemma take_perm (f : {ffun 'I_n -> A}) i (N : i < n):
  perm (take i.+1 (fgraph f)) (f (Ordinal N) :: take i (fgraph f)).
Proof.
set v := f (Ordinal N).
have X: perm (v :: take i (fgraph f)) (take i (fgraph f) ++ [:: v]).
- case:(perm_cons_cat_cons v (take i (fgraph f)) [::] (take i (fgraph f)))=>_ H.
  by apply: H; rewrite cats0.
move/perm_sym: X; apply: perm_trans.
suff: take i.+1 (fgraph f) = take i (fgraph f) ++ [:: v].
- move=>->. apply: perm_refl.
rewrite /f' !fgraph_codom /= !codomE -map_take. 
have N' : i < size (enum 'I_n) by rewrite size_enum_ord. 
rewrite (@take_plus1 _ i (enum 'I_n) (Ordinal N) N') map_cat /=.
rewrite /v map_take; congr (_ ++ _); congr (_ :: _); congr (f _).
by move: (@nth_ord_enum n (Ordinal N) (Ordinal N)). 
Qed.

Lemma take_eta_perm (f : {ffun 'I_n -> A}) i v (N : i < n):
  perm (take i.+1 (fgraph [ffun y => [eta f with Ordinal N |-> v] y])) 
       (v :: take i (fgraph f)).
Proof.
set f' := (fgraph [ffun y => [eta f with Ordinal N |-> v] y]).
have X: perm (v :: take i (fgraph f)) (take i (fgraph f) ++ [:: v]).
- case:(perm_cons_cat_cons v (take i (fgraph f)) [::] (take i (fgraph f)))=>_ H.
  by apply: H; rewrite cats0.
move/perm_sym: X.
apply: perm_trans.
suff: take i.+1 f' = take i (fgraph f) ++ [:: v].
- move=>->. apply: perm_refl.
rewrite /f' !fgraph_codom /= !codomE -map_take. 
have N' : i < size (enum 'I_n) by rewrite size_enum_ord. 
rewrite (@take_plus1 _ i (enum 'I_n) (Ordinal N) N') map_cat /=.
rewrite {2}ffunE /= {2}/eq_op /= nth_enum_ord // eq_refl. 
congr (_ ++ _).
rewrite -map_take; apply/eq_in_map; case=>x H.  
rewrite -(mem_map (@ord_inj n)) /= map_take val_enum_ord.
rewrite -{1}(subnKC (ltnW N)) iota_add take_cat size_iota ltnn /=. 
rewrite subnn take0 cats0 mem_iota ffunE /= /eq_op /=. 
by case: ltngtP.
Qed.

(********************************************************************)
(*                         Consumer Procedure                       *)
(********************************************************************)

Let consumer_post Ac := 
  fun (_ : unit) m => 
      let: H:= tb_self m in
      exists f,
      [/\ is_array Ac f (pv_self m), 
      pushed H [::] & popped H (fgraph f)].

Definition consumer_tp (Ac : {array [finType of 'I_n] -> A}) := 
  forall j,
  STsep [W] (fun i => 
               let: H:= tb_self i in
                exists f,
                [/\ is_array Ac f (pv_self i), j <= n,
                    pushed H [::] & popped H (take j (fgraph f))],
              fun (_ : unit) m => consumer_post Ac tt m).

Program Definition consumer Ac : consumer_tp Ac :=   
  ffix (fun (loop : consumer_tp Ac) i =>
    Do (if i == n 
        then ret tt
        else t <-- [pop_framed T];
             if t is Some v then
               [priv_write pv (Ac .+ i) v];;
               loop (i.+1)
             else loop i)).

Next Obligation. by apply: entangle_extR. Defined.
Next Obligation. by apply: entangle_extL. Defined.

Next Obligation. 
move=>i1 /=[f][[Ai]Vi] Nl Hpu Hpo; rewrite leq_eqVlt in Nl.
case: ifP Nl=>/= Nl N1.
- move/eqP: Nl=>X. subst i. apply:val_ret=>i2 M1; exists f.
  rewrite -(menvs_loc M1).
  have X: size (fgraph f) = n by rewrite fgraph_codom /= codomE size_map size_enum_ord.
  by rewrite -{1}X take_size in Hpo.  
move=>Ci1; move:(Ci1); apply:step.
case: (Ci1)=>_[ps][ts][Ei] Cp1 Ct1; subst i1.
rewrite (getC Cp1 Ci1) in Ai Vi; rewrite (getC Ct1 Ci1) in Hpu Hpo.
rewrite joinC; apply:val_extend=>//=.
apply: (gh_ex Unit); apply: (gh_ex (tb_self ts)); apply:val_do.
- by split=>//; apply:hist_unit (cohVSO Ct1).
move=>v j2[_]. 
case: v=>[v [t][ls][L2]Es2 Ct2 p2 Ci2 M2|[t][ls] Es2 Mt2 Ct2 p2 Ci2 M2]; first last.
- apply:val_do=>// _; exists f; move/menvs_coh: (M2)=>[_]Cp2.
  rewrite (getC Ct2 Ci2) (getC Cp2 Ci2) -(menvs_loc M2) Es2. 
  by split=>//; move/ltnW: N1.
apply: step; rewrite joinC; apply: val_extend; first by move/menvs_coh: M2=>[_].
move:(updi_split Ac (Ordinal N1) f)=> /= E; rewrite -Ai in E.
set hl := (updi _ _) in E; set hr := (updi _ _) in E.
apply: (gh_ex (hl \+ hr)); apply: val_do=>Cp2. 
- rewrite -(menvs_loc M2).
  rewrite (joinC hl) -joinA in E. exists A, (f (Ordinal N1)). rewrite E. 
  by rewrite /indx index_enum_ord /=.
move=>p3 Ep3 Cp3 st3 Ci3 M3. 
apply:val_do=>[_|_ i4 [f'][[Ai4]Vi4]Hpu4 Hpo4 Ci4 _]; last by exists f'.
move/menvs_coh:(M3)=>[_]Ct3.
set f' := [ffun y => [eta f with (Ordinal N1) |-> v] y].
exists f'; rewrite (getC Cp3 Ci3)(getC Ct3 Ci3) -(menvs_loc M3) Es2; split=>//.
rewrite Ep3; split; last first.
- by rewrite joinAC joinC in E; rewrite ?E !hvalidPtUn /indx index_enum_ord /= in Vi *.
- move:(updi_split Ac (Ordinal N1) f')=>/=->. 
  rewrite /f' /hl /hr !takeord !dropord /indx !index_enum_ord /= ffunE eq_refl /=.
  by rewrite -(joinC (Ac.+i :->_)) !joinA.
- move=>t' s1 s2.
  rewrite findUnL; last by move:(validL (cohVSO Ct2)); rewrite Es2.
  case:ifP=>[_|_ G]; first by move:(Hpu t' s1 s2).
  move/find_some: (G); rewrite um_domPt inE /=; move/eqP=>Z; subst t'.
  by rewrite um_findPt in G; case: G=><-<-; right; exists v.
rewrite /f' /=; apply: (popped_perm (xs := v :: take i (fgraph f))); first last.
by apply/perm_sym; apply: take_eta_perm. 
by exists t, ls, (tb_self ts); rewrite joinC.
Qed.

(********************************************************************)
(*                         Producer Procedure                       *)
(********************************************************************)

Variables (al : nat) (asent alock : ptr).
Notation Alloc := (AllocCAS al asent alock). 
Notation V := (entangle W Alloc). 
Notation al_coh := (al_coh al asent alock). 

Let producer_post Ap f := 
  fun (_ : unit) m => 
      let: H:= tb_self m in
      [/\ is_array Ap f (pv_self m), 
      pushed H (fgraph f) & popped H [::]].

Definition producer_tp (Ap : {array [finType of 'I_n] -> A}) := 
 forall j,
 {f}, STsep [V] (fun i => 
            let: H:= tb_self i in
             [/\ is_array Ap f (pv_self i), j <= n,
                 pushed H (take j (fgraph f)) & popped H [::]],
            fun (_ : unit) m => producer_post Ap f tt m).

Program Definition producer Ap : producer_tp Ap :=   
  ffix (fun (loop : producer_tp Ap) i =>
    Do (if i == n 
        then ret tt
        else e <-- [priv_read pv A (Ap .+ i)];
             push_framed pv al asent alock e;;
             loop (i.+1))).

Next Obligation. exact: injLE (entangle_apartA _ _ _). Defined.

Next Obligation.
apply: gh=>i1 f [Ep1] N1 Hpu1 Hpo1; case:ifP=>Z; case:(Ep1)=>Ai1 V1 Ci1.
- move/eqP: Z=>Z; subst i. 
  apply:val_ret =>i2 M1; rewrite /producer_post -!(menvs_loc M1).
  have X: size (fgraph f) = n by rewrite fgraph_codom /= codomE size_map size_enum_ord.
  rewrite -{1}X take_size in Hpu1.  
  by case: Ep1=> G1 G2. 
apply: step; case: (Ci1)=>_ [pts1][as1][Zs]Cpt1 Ca1;subst i1. 
case: Cpt1=>_[sp1][st1][Zs]Cp1 Ct1; subst pts1.
rewrite (getC Ct1 Ci1) /= in Hpu1 Hpo1; rewrite (getC Cp1 Ci1) /= in Ep1.
rewrite -joinA; apply: val_extend=>//; rewrite leq_eqVlt Z /= in N1.
(* Now we need to construct existential from the array *)
move:(updi_split Ap (Ordinal N1) f)=> /= E; rewrite -Ai1 in E.
set hl := (updi _ _) in E; set hr := (updi _ _) in E.
rewrite (joinC hl) -[_ \+ hl \+ _]joinA in E.
apply: (gh_ex (f (Ordinal N1))); apply: (gh_ex (hl \+ hr)). 
rewrite (getC Cp1 Ci1) /indx index_enum_ord /= in E; apply: val_do=>//. 
move=>v sp2[->{v}] Ep2 Cp2 sta2 Ci2 M2. 
case/menvs_coh: (M2)=>_ [_ [st2][sa2][Z2]] Ct2 Ca2; subst sta2.
rewrite !joinA in Ci2 *.
apply:step; apply:(gh_ex Unit). 
apply:(gh_ex (tb_self st1)); apply:(gh_ex (pv_self sp1)).
apply: val_do=>[_|_ i3[t][ls]]; rewrite ?(getC Ct2 Ci2) ?(getC Cp2 Ci2)/=. 
have X: tb_self st2 = tb_self st1.
- case:(menvs_split (apart_extL _ _) Ct1 Ct2 M2)=>Mt2 _. 
  by rewrite -!(menvs_loc Mt2).
rewrite X; split=>//.
  by rewrite Ep2 E.
  by apply:hist_unit; move /cohVSO: Ct2=>/=; rewrite X.  
case=>Es3 Ep3 U3 L Ci3.
apply: (gh_ex f); apply: val_do=>[_|_]//.
rewrite Es3 /=; split=>//; first by rewrite Ep3.
- apply: (pushed_perm (xs := (f (Ordinal N1))::take i (fgraph f))); first last.
  apply/perm_sym; apply: take_perm. simpl.
  by exists t, ls, (tb_self st1); rewrite joinC. 
case:(Ci3)=>_[spt3][sa3][Z3]Cpt3 _; subst i3.
case:Cpt3=>_[sp3][st3][Z3]Cp3 Ct3; subst spt3.
rewrite ?(getC Ct3 Ci3) ?(getC Cp3 Ci3) /= in Es3 Ep3 U3.
move/cohVSO: Ct3 =>/validL /=; rewrite Es3 => V3.
move=>t' s1 s2; case X: (t' == t).
- move/eqP: X=>->{t'}. 
  rewrite findUnR // um_domPt inE eq_refl /= um_findPt. 
  by case=><-<-; right; eexists (f (Ordinal N1)). 
by rewrite findUnR // um_domPt inE eq_sym X /=; apply:Hpo1.
Qed.

(*******************************************

Main producer/consumer procedure:

- Ap - pointer to the producer array (of size n);
- Ac - pointer to the initial consumer array (of size n);
- ssent - a pointer to the exchange stack sentinel.

********************************************)

Notation U := (entangle (Priv pv) Alloc). 

(* Constructing parallel composition of producer/consumer *)
Definition prodcons_parcomp_tp (Ap Ac: {array [finType of 'I_n] -> A}):=
 {fp}, STsep [V] 
     (fun i => exists ps cs,
               [/\ tb_self i = 0 \\-> ([::], [::]), pv_self i = ps \+ cs, 
                   is_array Ap fp ps & exists fc, is_array Ac fc cs],
     fun (_ : unit) m => 
               let: H := tb_self m in
               exists ps cs fc, 
               [/\ pv_self m = ps \+ cs, 
                   is_array Ap fp ps, is_array Ac fc cs,
                   pushed H (fgraph fp) & popped H (fgraph fc)]).

Program Definition prodcons_parcomp Ap Ac : prodcons_parcomp_tp Ap Ac := 
  Do (par (producer Ap 0) [consumer Ac 0];; 
      ret tt).
Next Obligation. by exact: injLE (erefl _). Defined.
Next Obligation. 
apply: gh=>i fp [ps][cs][St Sp [Ep Vp] [fc][Ec Vc]] C; apply: step.
case: C (C) St Sp=>_ [x][xa][->{i}] Cpt. 
case: Cpt (Cpt)=>_ [xp][xt][->{x}] Cp Ct Cpt Ca C St Sp.
rewrite (getC Ct C) /= in St; rewrite (getC Cp C) /= in Sp.
set gcs := [:: priv_getcoh pv : getcoh; tb_getcoh T; al_getcoh al asent alock].
have E : sself gcs (xp \+ xt \+ xa) = 
         (ps, (0 \\-> ([::], [::]), self (alg al) xa)) \+ (cs, (Unit, Unit)).
- by rewrite /= (getC Cp C) (getC Ct C) (getC Ca C) pcm_joinE /= unitR Sp St.  
apply: (par_do (r1:=producer_post Ap fp) (r2:=consumer_post Ac)) E _ _ _.
- by rewrite /= starA. 
- apply: (gh_ex fp); apply: val_do=>// C'; rewrite !getV ?(cohV C') //= take0.
  by split=>// j s1 s2 /um_findPt_inv [->][<-<-]; left. 
- rewrite /= joinA; apply: vrf_coh=>C'; apply: val_extend.
  - apply/(star_coh_prec C'); rewrite !(getC Ca C) /= fsE -zipPt zipC.
    by rewrite -zagE -coh_zigzag zigE zipPt unitR -fsE -(cohE Ca). 
  apply: val_do.
  - move=>Cpt'; exists fc; rewrite !getV ?(cohV Cpt') //= take0.
    by split=>// j s1 s2; rewrite find0E. 
  (* it's annoying here to have to destruct C''; getters should be *)
  (* generalized so that a getter for * is a union of individual getters *)
  case=>m [fc' H] C'' j' Cmj' M; exists fc'.
  case: C'' (C'') H Cmj'=>_ [xp'][xt'][->] Cp' Ct' C'' H Cmj'.
  rewrite (getC Cp' C'') (getC Ct' C'') in H.
  by rewrite (getC Cp' Cmj') (getC Ct' Cmj'). 
move=>v x x1 x2 C' C1' C2' S' Hp [f' Hc].
apply: val_ret=>m M; exists x1.1, x2.1, f'.
move: S'; rewrite sselfE {M}(menvs_loc M) -sselfE; case=>->-> _.
case: Hp Hc; rewrite !getV ?(cohV C1') ?(cohV C2') //=. 
case=>->-> Hp1 Hp2 [[->->]] Hc1 Hc2; split=>//.
- by apply: pushed_mono Hp1 Hc1.
by rewrite joinC; apply: popped_mono Hc2 Hp2.
Qed.
  
Definition producer_consumer_tp (Ap Ac: {array [finType of 'I_n] -> A}) := 
 {fp}, STsep [U] (fun i => 
            exists ip ic fc, 
             [/\ l \notin sdom i, 
                 pv_self i = sent :-> null \+ ip \+ ic, 
                 is_array Ap fp ip & is_array Ac fc ic],
            fun (_ : unit) m => 
            exists mp mc mst h (p : ptr) fc, 
             [/\ pv_self m = mp \+ mc \+ mst, 
                 is_array Ap fp mp, is_array Ac fc mc, 
                 mst = sent :-> p \+ h & 
                 perm (fgraph fp) (fgraph fc)]).

Program Definition producer_consumer Ap Ac : 
  producer_consumer_tp Ap Ac := 
  Do (priv_hide (TreiberDecorate T) (Unit, Unit) 
                (cast (T:=fun m => ST m _) (entangleAC _ _ _) 
                      (prodcons_parcomp Ap Ac))).

Next Obligation.
apply:gh=>i1 fp[ip][ic][fc][Nd]Ep1[Ep Vp][Ec Vc] Ci1.
case: (Ci1)=>_[sp1][sa1][Zi1]Cp1 Ca1; subst i1; rewrite (getC Cp1 Ci1) /= in Ep1.
case: (Cp1)=>_[hL][hE]Zp1; rewrite Zp1 getE -joinA in Ep1 *.
pose st1 := l ->> [(0 \\-> ([::], [::]): hist A), sent :-> null, Unit].
have X: flats st1 = sent :-> null by rewrite /st1 tb_flatsE.
rewrite -X joinC in Ep1; rewrite Ep1.
have V:  valid (pv ->> [ip \+ ic, tt, hE] \+ sa1 \+ st1).
- rewrite /st1; apply:validSt. 
  by apply: (validPvSubst (hL:=hL)); move/cohV: Ci1; rewrite Zp1.
  by apply: (notinPvSubst (hL:=hL)); rewrite -Zp1. 
apply: val_privhide'=>//.
- exists Unit, [::], [::], null; rewrite !unitR; split=>//.
  exists null, Unit, Unit, [::]; rewrite !getE !unitR; split=>//.
  - by rewrite last_keyPt um_findPt; exists [::]. 
  - by split; [apply: contPt | apply: complete0].
  by move/cohH: (Cp1); rewrite Zp1 /validH priv_flatsE Ep1 X; move/validL /validR.
apply:val_cast; apply: (gh_ex fp); apply:val_do=>[_|_ m [ps][cs][fc']].
- by exists ip, ic; rewrite !getV //; split=>//; exists fc. 
case=> [Epv][Eps] Vps[Ecs]Vcs Hpu Hpo Cm i' j' hL' hE' g' Em Td.
exists ps, cs; case: Td=>h' [ls][ls'][p]/=[[Ej'] L'] Ct'. 
exists  (flats j'), (h' \+ g'.2), p, fc'.
rewrite getE /= Ej' tb_flatsE /= -joinA; last first. 
- by move/cohV: Cm; subst m; move/validL; apply: (validPvSubst (hL:=hL')).
split=>//.
- congr (_ \+ _); rewrite Em -joinA getE //= in Epv. 
  by rewrite joinA; move/cohV: Cm; rewrite Em.
rewrite Em Ej' joinC getE /= in Hpu Hpo; last first.
- by move/cohV: Cm; subst m j'; rewrite joinC. 
case: Ct'=>p'[h1][g1][ls'']. 
rewrite Ej' !getE /= unitR=>[[_]]Eh L'' Hlast [Hcont Hcom] _.
case: (push_pop_all Hcom Hcont Hpu Hpo)=>L[Ins]Hperm.
have Y: size (fgraph fp) = size (fgraph fc').
- by rewrite !fgraph_codom /= !codomE 2!size_map !size_enum_ord.
move/perm_size: (Hperm); rewrite Y size_cat.
set m' := (size (fgraph fc')).
case G: (size L)=>[|n']; last first.
- move: (ltnSn m')=>H1 Z; rewrite {1}Z addnS in H1. 
  move:(leq_addr n' m') (ltnSn (m' + n'))=> H2 H3.
  by move:(leq_ltn_trans H2 H3); rewrite {1}Z addnS ltnn.
by move/size0nil: G=> Zl; subst L; rewrite cats0 in Hperm. 
Qed.

End ClientPrograms.






